def get_logger(filename):
    """
    :param project_path: path_root
    :param filename: file.txt
    :return:
    """
    import logging

    if not filename:
        raise Exception('require var filename: {file_path.txt} ')

    logging.basicConfig(
        filename=filename,
        format="[%(levelname)s]%(asctime)s:%(name)s:%(message)s",
        datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger('main')
    logger.setLevel(logging.INFO)
    return logger



# bq.py
txt_set_project_success = "Project update: {0}"
txt_set_project_errors = "Project Not Found: {0}"

txt_execute_sql_success = "Success file: {0}"
txt_execute_sql_errors = "Error file: {0}"

# bigquery.py
txt_create_dt_success = "Created dataset: {0}"
txt_create_dt_error = "Already Exists or ProjectID Not Found: {0}"

txt_delete_dt_success = "Deleted dataset: {0}"
txt_delete_dt_error = "Dataset Not Found: {0}"

txt_create_tb_success = "Created table: {0}"
txt_create_tb_error = "Already Exists or TableID Not Found: {0}"

txt_delete_tb_success = "Deleted table: {0}"
txt_delete_tb_error = "Not exists Table or TableID Not Found: {0}"

txt_query_to_df_success = "Success query to dataframe"
txt_query_to_df_errors = "Error query to dataframe"

txt_query_to_rows_success = "Success query to rows"
txt_query_to_rows_errors = "Error query to rows"

txt_truncate_success = "Success truncate query"
txt_truncate_errors = "Error truncate query"

txt_delete_partition_success = "Success delete range partition {0} to {1}"
txt_delete_partition_errors = "Error delete range partition {0} to {1}"

txt_restore_table_success = "Success restore table name {0}_RESTORE"
txt_restore_table_errors = "Error restore table name {0}_RESTORE"

# dataframe
txt_upload_df_to_bq_success = "Load Table {0} from dataframe"
txt_upload_df_to_bq_errors = "Provided Schema does not match table {0}"

# cloudstorage
txt_gs_bucket_list_patterns_success = "list_blobs from  bucket {0}"
txt_gs_bucket_list_patterns_errors = "BucketName Not Found {0}"

txt_gs_bucket_list_to_df_success = "blobs from dataframe {0}/{1}"
txt_gs_bucket_list_to_df_errors = "BucketName Not Found {0} "

txt_bucket_download_file_success = "Download file: {0}"
txt_bucket_download_file_errors = "File Not Found {0}"

txt_bucket_upload_file_success = "Download file: {0}"
txt_bucket_upload_file_errors = "File Not Found {0}"

txt_storage_delete_file_success = "Delete file: {0}"
txt_storage_delete_file_errors = "File Not Found {0}"

txt_storage_delete_folder_success = "Delete folder: {0}"
txt_storage_delete_folder_errors = "Folder Not Found {0}"

# dataset
txt_show_describe_table_success = "Success describe table {0}"
txt_show_describe_table_errors = "Error describe table {0}"



from etl_bq_tools import bq_query_to_df
from etl_bq_tools.utils import constants as cons
from etl_bq_tools.utils.prettytable import format_for_print
from etl_bq_tools.utils.time_execution import get_time_function_execution


@get_time_function_execution
def show_describe_dataset(project_id=None, dataset_full_id="",
                          logging=None, key_path=None, show_table=False):
    """
     :param project_id: String
     :param dataset_full_id: String {project}.{dateset}
     :param logging: object
     :param key_path: file.json
     :param show_table: False
    """
    from color_tools import cprint
    from sizebytes_tools import convert_bytes

    query_text = """
        SELECT
          project_id                                                             AS project_id
        , dataset_id                                                             AS dataset_id
        , table_id                                                               AS table_id
        , size_bytes                                                             AS size_bytes
        , TIMESTAMP_MILLIS(creation_time)                                        AS creation_time
        , TIMESTAMP_MILLIS(last_modified_time)                                   AS last_modified_time
        , CAST(DATETIME(TIMESTAMP_MILLIS(last_modified_time), "America/Lima") AS DATETIME) AS  last_modified_datetime_local
        , CAST(DATETIME(TIMESTAMP_MILLIS(last_modified_time), "America/Lima") AS DATE)     AS  last_modified_date_local
        , EXTRACT(HOUR FROM CAST(DATETIME(TIMESTAMP_MILLIS(last_modified_time), "America/Lima") AS DATETIME)) AS last_modified_hour_local
        , row_count
        , CASE
            WHEN type = 1
              THEN 'TABLE'
            WHEN type = 2
              THEN 'VIEW'
            ELSE NULL
          END                                                                     AS table_type
        FROM
          `{0}`.__TABLES__
    ;
    """

    try:
        _sql = query_text.format(dataset_full_id)
        df = bq_query_to_df(
            project_id=project_id,
            sql=_sql,
            file=None,
            use_legacy_sql=False,
            key_path=key_path
        )
        df["creation_time"] = df["creation_time"].astype(str).str[:19]
        df["last_modified_time"] = df["last_modified_time"].astype(str).str[:19]
        df["last_modified_datetime_local"] = df["last_modified_datetime_local"].astype(str).str[:19]
        df["size_bytes_format"] = df.apply(lambda x: convert_bytes(x["size_bytes"]), axis=1)

        if show_table:
            print(format_for_print(df))

        if logging:
            logging.info(cons.txt_show_describe_table_success.format(f"{dataset_full_id}"))
            cprint(cons.txt_show_describe_table_success.format(f"{dataset_full_id}"), "yellow")
        else:
            cprint(cons.txt_show_describe_table_success.format(f"{dataset_full_id}"), "yellow")
        return df
    except:
        if logging:
            logging.info(cons.txt_show_describe_table_errors.format(f"{dataset_full_id}"))
            cprint(cons.txt_show_describe_table_errors.format(f"{dataset_full_id}"), "yellow")
        else:
            cprint(cons.txt_show_describe_table_errors.format(f"{dataset_full_id}"), "yellow")



from etl_bq_tools.utils.time_execution import get_time_function_execution
from etl_bq_tools.utils import constants as cons


def bq_set_project(project_id=None, logging=None):
    """
     :param project_id: String
     :param logging: Object
     :return:
    """
    import os
    from color_tools import cprint

    if not project_id:
        raise Exception('require var project_id: {project_id} ')

    try:
        os.system(f"gcloud config set project {project_id}")

        if logging:
            logging.info(cons.txt_set_project_success.format(project_id))
            cprint(cons.txt_set_project_success.format(project_id), 'yellow')
        else:
            cprint(cons.txt_set_project_success.format(project_id), 'yellow')
    except:
        if logging:
            logging.info(cons.txt_set_project_errors.format(project_id))
            cprint(cons.txt_set_project_errors.format(project_id), 'yellow')
        else:
            cprint(cons.txt_set_project_errors.format(project_id), 'yellow')


@get_time_function_execution
def bq_execute_sql(sql=None, parameter=None, logging=None):
    """
     :param sql: file.sql
     :param parameter: list dict params [{"variable": "", "type": "", "value":""}]
     :param logging: object
     :return:
    """

    import os
    from color_tools import cprint

    if not sql:
        raise Exception('require var sql: {file.sql} ')

    file_sql = os.path.basename(sql)

    try:
        _parameter2 = ""
        if parameter:
            parameter = {"parameter": parameter}
            list_params = [params_values for _, params_values in parameter.items()][0]
            for value in list_params:
                _var = value["variable"]
                _type = value["type"]
                _val = value["value"]
                _parameter = f"--parameter={_var}:{_type}:{_val} "
                _parameter2 += f"{_parameter}"
            sentence_bq = f"bq query --quiet  --use_legacy_sql=false {_parameter2} --flagfile={sql} "
        else:
            sentence_bq = f"bq query --quiet  --use_legacy_sql=false --flagfile={sql} "

        os.system(sentence_bq)

        if logging:
            logging.info(cons.txt_execute_sql_success.format(file_sql))
            cprint(cons.txt_execute_sql_success.format(file_sql), 'yellow')
        else:
            cprint(cons.txt_execute_sql_success.format(file_sql), 'yellow')

    except Exception as e:
        if logging:
            logging.info(cons.txt_execute_sql_errors.format(file_sql))
            cprint(cons.txt_execute_sql_errors.format(file_sql), 'yellow')
        else:
            cprint(cons.txt_execute_sql_errors.format(file_sql), 'yellow')



from etl_bq_tools.utils.time_execution import get_time_function_execution
from etl_bq_tools.utils import constants as cons


@get_time_function_execution
def bq_create_dataset(project_id=None,
                      dataset_id=None,
                      logging=None,
                      key_path=None):
    """
    :param project_id: String
    :param dataset_id: String
    :param key_path: file.json
    :param logging: object
    :return:
    """
    from google.cloud import bigquery
    from google.api_core.exceptions import Conflict
    from color_tools import cprint

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not dataset_id:
        raise Exception('require var dataset_id:{dataset_id} ')

    client = bigquery.Client(project=project_id)
    if project_id and key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    dataset_id = f"{project_id}.{dataset_id}"
    dataset = bigquery.Dataset(dataset_id)
    dataset.location = "US"

    try:
        dataset = client.create_dataset(dataset, exists_ok=True)
        if logging:
            logging.info(cons.txt_create_dt_success.forma(f"{project_id}.{dataset.dataset_id}"))
            cprint(cons.txt_create_dt_success.forma(f"{project_id}.{dataset.dataset_id}"), "yellow")
        else:
            cprint(cons.txt_create_dt_success.forma(f"{project_id}.{dataset.dataset_id}"), "yellow")
    except Conflict:
        if logging:
            logging.info(cons.txt_create_dt_error.forma(f"{project_id}.{dataset.dataset_id}"))
            cprint(cons.txt_create_dt_error.format(f"{project_id}.{dataset.dataset_id}"), "yellow")
        else:
            cprint(cons.txt_create_dt_error.format(f"{project_id}.{dataset.dataset_id}"), "yellow")


@get_time_function_execution
def bq_delete_dataset(project_id=None,
                      dataset_id=None,
                      logging=None,
                      key_path=None):
    """
    :param project_id: String
    :param dataset_id: String
    :param logging: object
    :param key_path: file.json
    :return:
    """
    from google.cloud import bigquery
    from google.cloud.exceptions import Conflict
    from color_tools import cprint

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not dataset_id:
        raise Exception('require var dataset_id:{dataset_id} ')

    client = bigquery.Client(project=project_id)
    if project_id and key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    dataset_id = f"{project_id}.{dataset_id}"

    try:
        client.delete_dataset(dataset_id, delete_contents=True, not_found_ok=True)
        if logging:
            logging.info(cons.txt_delete_dt_success.format(dataset_id))
            cprint(cons.txt_delete_dt_success.format(dataset_id), "yellow")
        else:
            cprint(cons.txt_delete_dt_success.format(dataset_id), "yellow")
    except Conflict:
        if logging:
            logging.info(cons.txt_delete_dt_error.format(f"{project_id}.{dataset_id}"))
            cprint(cons.txt_delete_dt_error.format(f"{project_id}.{dataset_id}"), "yellow")
        else:
            cprint(cons.txt_delete_dt_error.format(f"{project_id}.{dataset_id}"), "yellow")


@get_time_function_execution
def bq_list_dataset(project_id=None,
                    logging=None,
                    key_path=None):
    """
    :param project_id: String
    :param logging: object
    :param key_path: file.json
    :return:
    """
    from google.cloud import bigquery

    client = bigquery.Client(project=project_id)
    if project_id and key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    datasets = list(client.list_datasets())

    data_list = list()
    if datasets:
        logging.info(f"Datasets in project {project_id}:")
        for dataset in datasets:
            data_dict = dict()
            data_dict["project_id"] = dataset.project
            data_dict["dataset_id"] = dataset.dataset_id
            data_list.append(data_dict)
            print(f"\t {dataset.project}.{dataset.dataset_id}")
    else:
        logging.info(f"{project_id} project does not contain any datasets.")

    return data_list


@get_time_function_execution
def bq_create_table(project_id=None,
                    dataset_id=None,
                    table_id=None,
                    schema=None,
                    logging=None,
                    key_path=None):
    """
    :param project_id: String
    :param dataset_id: String
    :param table_id: String
    :param schema: dict
    :param logging: object
    :param key_path: file.json
    :return:
    """
    from google.cloud import bigquery
    from google.api_core.exceptions import Conflict
    from color_tools import cprint

    if not project_id:
        raise Exception('require var project_id: {project_id} ')
    if not dataset_id:
        raise Exception('require var dataset_id: {dataset_id} ')
    if not table_id:
        raise Exception('require var table_id: {table_id} ')
    if not schema:
        raise Exception('require var schema ')

    client = bigquery.Client(project=project_id)
    if project_id and key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    table_id = f"{project_id}.{dataset_id}.{table_id}"
    table = bigquery.Table(table_id, schema=schema)

    try:
        table = client.create_table(table)
        if logging:
            logging.info(cons.txt_create_tb_success.format(f"{table.project}.{table.dataset_id}.{table.table_id}"))
            cprint(cons.txt_create_tb_success.format(f"{table.project}.{table.dataset_id}.{table.table_id}"), "yellow")
        else:
            cprint(cons.txt_create_tb_success.format(f"{table.project}.{table.dataset_id}.{table.table_id}"), "yellow")
    except Conflict:
        if logging:
            logging.info(cons.txt_create_tb_error.format(f"{table.project}.{table.dataset_id}.{table.table_id}"))
            cprint(cons.txt_create_tb_error.format(f"{table.project}.{table.dataset_id}.{table.table_id}"), "yellow")
        else:
            cprint(cons.txt_create_tb_error.format(f"{table.project}.{table.dataset_id}.{table.table_id}"), "yellow")


@get_time_function_execution
def bq_delete_table(project_id=None,
                    table_full_id=None,
                    logging=None,
                    key_path=None):
    """
    :param project_id: String
    :param table_full_id: String -> 'project.dataset.table'
    :param logging: object
    :param key_path: file.json
    :return:
    """

    from google.cloud import bigquery
    from google.api_core.exceptions import Conflict
    from color_tools import cprint

    client = bigquery.Client(project=project_id)
    if project_id and key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    table_id = f"{table_full_id}"
    try:
        client.delete_table(table_id, not_found_ok=True)
        if logging:
            logging.info(cons.txt_delete_tb_success.format(f"{table_id}"))
            cprint(cons.txt_delete_tb_success.format(f"{table_id}"), "yellow")
        else:
            cprint(cons.txt_delete_tb_success.format(f"{table_id}"), "yellow")
    except Conflict:
        if logging:
            logging.info(cons.txt_delete_tb_error.format(f"{table_id}"))
            cprint(cons.txt_delete_tb_error.format(f"{table_id}"), "yellow")
        else:
            cprint(cons.txt_delete_tb_error.format(f"{table_id}"), "yellow")


@get_time_function_execution
def bq_query_to_df(
        project_id=None, sql=None, file=None, use_legacy_sql=False,
        logging=None, key_path=None):
    """
     :param project_id: String
     :param sql: String
     :param file: String file_path.sql
     :param use_legacy_sql: Boolean
     :param logging: object
     :param key_path: file.json
     :return:
    """
    from google.cloud import bigquery
    from color_tools import cprint

    if not project_id:
        raise Exception('require var project_id: {project_id} ')
    if not sql and not file:
        raise Exception('Required var sql or file')
    if sql and file:
        raise Exception('Only one variable is required {sql} or {file}')
    if file:
        if not str(file).endswith(".sql"):
            raise Exception('the file only support extension {file_path}.sql')

    client = bigquery.Client(project=project_id)
    if key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    sql_text = ""
    if sql and file is None:
        sql_text = sql

    if file and sql is None:
        sql_text = open(file).read()

    try:
        query_config = bigquery.QueryJobConfig(use_legacy_sql=use_legacy_sql)
        query_job = client.query(sql_text, job_config=query_config)
        df = query_job.to_dataframe()
        if logging:
            logging.info(cons.txt_query_to_df_success)
            cprint(cons.txt_query_to_df_success, 'yellow')
        else:
            cprint(cons.txt_query_to_df_success, 'yellow')
        return df
    except Exception as e:
        if logging:
            logging.info(cons.txt_query_to_df_errors)
            cprint(cons.txt_query_to_df_errors, 'yellow')
        else:
            cprint(cons.txt_query_to_df_errors, 'yellow')


@get_time_function_execution
def bq_query_to_rows(
        project_id=None, sql=None, file=None, use_legacy_sql=False,
        logging=None, key_path=None):
    """
     :param project_id: String
     :param sql: String
     :param file: String file_path.sql
     :param use_legacy_sql: Boolean
     :param logging: object
     :param key_path: file.json
     :return:
    """
    from google.cloud import bigquery
    from color_tools import cprint

    if not project_id:
        raise Exception('require var project_id: {project_id} ')
    if not sql and not file:
        raise Exception('Required var sql or file')
    if sql and file:
        raise Exception('Only one variable is required {sql} or {file}')
    if file:
        if not str(file).endswith(".sql"):
            raise Exception('the file only support extension {file_path}.sql')

    client = bigquery.Client(project=project_id)
    if key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    sql_text = ""
    if sql and file is None:
        sql_text = sql

    if file and sql is None:
        sql_text = open(file).read()

    try:
        query_config = bigquery.QueryJobConfig(use_legacy_sql=use_legacy_sql)
        query_job = client.query(sql_text, job_config=query_config)
        rows = query_job.result()
        if logging:
            logging.info(cons.txt_query_to_rows_success)
            cprint(cons.txt_query_to_rows_success, 'yellow')
        else:
            cprint(cons.txt_query_to_rows_success, 'yellow')
        return rows
    except Exception as e:
        if logging:
            logging.info(cons.txt_query_to_rows_errors)
            cprint(cons.txt_query_to_rows_errors, 'yellow')
        else:
            cprint(cons.txt_query_to_rows_errors, 'yellow')


@get_time_function_execution
def bq_query_truncate(
        project_id=None, table_full_id="",
        logging=None, key_path=None):
    """
     :param project_id: String
     :param table_full_id: String -> 'project.dataset.table'
     :param logging: object
     :param key_path: file.json
     :return:
    """
    from google.cloud import bigquery
    from color_tools import cprint

    if not project_id:
        raise Exception('require var project_id: {project_id} ')
    if not table_full_id:
        raise Exception('require var table_full_id: {table_full_id} ')

    client = bigquery.Client(project=project_id)
    if key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    sql_text = f"""TRUNCATE TABLE `{table_full_id}` """
    try:
        query_config = bigquery.QueryJobConfig(use_legacy_sql=False)
        query_job = client.query(sql_text, job_config=query_config)
        rows = query_job.result()
        if logging:
            logging.info(cons.txt_truncate_success)
            cprint(cons.txt_truncate_success, 'yellow')
        else:
            cprint(cons.txt_truncate_success, 'yellow')
        return rows
    except Exception as e:
        if logging:
            logging.info(cons.txt_truncate_errors)
            cprint(cons.txt_truncate_errors, 'yellow')
        else:
            cprint(cons.txt_truncate_errors, 'yellow')


@get_time_function_execution
def bq_query_delete_partition(
        project_id=None, table_full_id="",
        partition_column="_PARTITIONTIME",
        date_initial=None, date_final=None,
        logging=None, key_path=None):
    """
     :param project_id: String
     :param table_full_id: String -> 'project.dataset.table'
     :param partition_column: String
     :param date_initial: String -> YYYY-MM-DD
     :param date_final: String -> YYYY-MM-DD
     :param logging: object
     :param key_path: file.json
     :return:
    """
    from google.cloud import bigquery
    from color_tools import cprint

    if not project_id:
        raise Exception('require var project_id: {project_id} ')
    if not table_full_id:
        raise Exception('require var table_full_id: {table_full_id} ')
    if not partition_column:
        raise Exception('require var partition_column: {partition_column} ')
    if not date_initial:
        raise Exception('require var date_initial: {date_initial} ')
    if not date_final:
        raise Exception('require var date_final: {date_final} ')

    client = bigquery.Client(project=project_id)
    if key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    sql_text = f"""DELETE FROM `{table_full_id}` 
                   WHERE DATE({partition_column}) between '{date_initial}' and '{date_final}' """

    try:
        query_config = bigquery.QueryJobConfig(use_legacy_sql=False)
        query_job = client.query(sql_text, job_config=query_config)
        rows = query_job.result()
        if logging:
            logging.info(cons.txt_delete_partition_success.format(f"{date_initial}", f"{date_final}"))
            cprint(cons.txt_delete_partition_success.format(f"{date_initial}", f"{date_final}"), 'yellow')
        else:
            cprint(cons.txt_delete_partition_success.format(f"{date_initial}", f"{date_final}"), 'yellow')
        return rows
    except Exception as e:
        if logging:
            logging.info(cons.txt_delete_partition_errors.format(f"{date_initial}", f"{date_final}"))
            cprint(cons.txt_delete_partition_errors.format(f"{date_initial}", f"{date_final}"), 'yellow')
        else:
            cprint(cons.txt_delete_partition_errors.format(f"{date_initial}", f"{date_final}"), 'yellow')


@get_time_function_execution
def bq_query_restore_table(
        project_id=None, table_full_id="", interval=1,
        logging=None, key_path=None):
    from google.cloud import bigquery
    from color_tools import cprint

    if not project_id:
        raise Exception('require var project_id: {project_id} ')
    if not table_full_id:
        raise Exception('require var table_full_id: {table_full_id} ')

    client = bigquery.Client(project=project_id)
    if key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    sql_text = f""" CREATE OR REPLACE TABLE `{table_full_id}_RESTORE` AS 
                    SELECT * FROM `{table_full_id}` FOR SYSTEM_TIME AS OF TIMESTAMP_SUB(CURRENT_TIMESTAMP(), INTERVAL {interval} HOUR)
                """

    try:
        query_config = bigquery.QueryJobConfig(use_legacy_sql=False)
        query_job = client.query(sql_text, job_config=query_config)
        rows = query_job.result()
        if logging:
            logging.info(cons.txt_restore_table_success.format(f"{table_full_id}"))
            cprint(cons.txt_restore_table_success.format(f"{table_full_id}"), 'yellow')
        else:
            cprint(cons.txt_restore_table_success.format(f"{table_full_id}"), 'yellow')
        return rows
    except Exception as e:
        if logging:
            logging.info(cons.txt_restore_table_errors.format(f"{table_full_id}"))
            cprint(cons.txt_restore_table_errors.format(f"{table_full_id}"), 'yellow')
        else:
            cprint(cons.txt_restore_table_errors.format(f"{table_full_id}"), 'yellow')



from etl_bq_tools.utils.time_execution import get_time_function_execution
from etl_bq_tools.utils import constants as cons


@get_time_function_execution
def upload_df_to_bq(
        project_id, table_full_id, key_path, partition_column=None,
        schema=None, df=None, if_exists="truncate", logging=None):
    """
     :param project_id: String
     :param table_full_id: String -> 'project.dataset.table'
     :param key_path: file.json
     :param partition_column: Date
     :param schema: dict
     :param df: Dataframe
     :param if_exists: String -> "append" or "truncate"
     :param logging: object
     :return:
    """
    from google.cloud import bigquery
    from google.api_core.exceptions import Conflict
    import pandas as pd
    from color_tools import cprint

    if not project_id:
        raise Exception('require var project_id: {project_id} ')
    if not table_full_id:
        raise Exception('require var table_full_id: {table_full_id} ')
    if not isinstance(df, pd.DataFrame):
        raise Exception('require var df: {df} ')
    if not if_exists:
        raise Exception('require var if_exists: {if_exists} ')

    client = bigquery.Client(project=project_id)
    if key_path:
        client = bigquery.Client.from_service_account_json(key_path)

    table_id = f"{table_full_id}"

    job_config = bigquery.LoadJobConfig()

    if partition_column:
        job_config.time_partitioning = bigquery.TimePartitioning(
            type_=bigquery.TimePartitioningType.DAY,
            field=partition_column,
            expiration_ms=None
        )

    if if_exists == "truncate":
        job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
    elif if_exists == "append":
        job_config.write_disposition = bigquery.WriteDisposition.WRITE_APPEND

    if schema:
        job_config.schema = schema
    else:
        job_config.autodetect = True

    try:
        job = client.load_table_from_dataframe(
            df, table_id, job_config=job_config
        )
        job.result()

        if logging:
            logging.info(cons.txt_upload_df_to_bq_success.format(table_id))
            cprint(cons.txt_upload_df_to_bq_success.format(table_id), 'yellow')
        else:
            cprint(cons.txt_upload_df_to_bq_success.format(table_id), 'yellow')
    except Conflict:
        if logging:
            logging.info(cons.txt_upload_df_to_bq_errors.format(table_id))
            cprint(cons.txt_upload_df_to_bq_errors.format(table_id), 'yellow')
        else:
            cprint(cons.txt_upload_df_to_bq_errors.format(table_id), 'yellow')



from etl_bq_tools.utils import constants as cons
from etl_bq_tools.utils.memory import get_reduce_memory


def gs_bucket_list_patterns(project_id, bucket_name,
                            prefix_start=None, prefix_end=None,
                            logging=None, key_path=None):
    """
    :param project_id: String
    :param bucket_name: String
    :param prefix_start: String
    :param prefix_end: String
    :param logging: Object
    :param key_path: file.json
    :return: list_blobs
    """
    from color_tools import cprint
    from google.cloud import storage

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not bucket_name:
        raise Exception('require var bucket_name:{bucket_name} ')

    client = storage.Client(project=project_id)
    if project_id and key_path:
        client = storage.Client.from_service_account_json(key_path)

    try:
        bucket = client.get_bucket(bucket_name)
        blobs = bucket.list_blobs()

        filenames = list()
        if prefix_start:
            filenames = list(sorted([str(blob.name) for blob in blobs
                                     if str(blob.name).startswith(prefix_start)]))

        elif prefix_end:
            filenames = list(sorted([str(blob.name) for blob in blobs
                                     if str(blob.name).endswith(prefix_end)]))

        elif prefix_start and prefix_start:
            filenames = list(sorted([str(blob.name) for blob in blobs
                                     if str(blob.name).startswith(prefix_start) and
                                     str(blob.name).endswith(prefix_end)]))

        elif not prefix_start and not prefix_start:
            filenames = list(sorted([str(blob.name) for blob in blobs]))

        if logging:
            logging.info(cons.txt_gs_bucket_list_patterns_success.format(bucket_name))
            cprint(cons.txt_gs_bucket_list_patterns_success.format(bucket_name), 'yellow')
        else:
            cprint(cons.txt_gs_bucket_list_patterns_success.format(bucket_name), 'yellow')
        return filenames
    except Exception as e:
        if logging:
            logging.info(cons.txt_gs_bucket_list_patterns_errors.format(bucket_name))
            cprint(cons.txt_gs_bucket_list_patterns_errors.format(bucket_name), 'yellow')
        else:
            cprint(cons.txt_gs_bucket_list_patterns_errors.format(bucket_name), 'yellow')


def gs_bucket_list_to_df(project_id, bucket_name,
                         filename=None, filename_col=None,
                         logging=None, key_path=None):
    import pandas as pd
    import gc
    from color_tools import cprint
    import gcsfs

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not bucket_name:
        raise Exception('require var bucket_name:{bucket_name} ')
    if not filename:
        raise Exception('require var filename:{filename} ')
    if not filename_col:
        raise Exception('require var filename_col:{filename_col} ')

    try:
        fs = gcsfs.GCSFileSystem(project=project_id,
                                 token=key_path)

        with fs.open(f"gs://{bucket_name}/{filename}") as f:
            frame = pd.read_csv(f)
            frame[f'{filename_col}'] = filename
            df_columns = frame.columns
            frame[df_columns] = frame[df_columns].astype(str)
            df2 = get_reduce_memory(frame, False)
            f.close()
        del frame
        gc.collect()

        if logging:
            logging.info(cons.txt_gs_bucket_list_to_df_success.format(bucket_name, filename))
            cprint(cons.txt_gs_bucket_list_to_df_success.format(bucket_name, filename), 'yellow')
        else:
            cprint(cons.txt_gs_bucket_list_to_df_success.format(bucket_name, filename), 'yellow')

        return df2
    except Exception as e:
        if logging:
            logging.info(cons.txt_gs_bucket_list_to_df_errors.format(bucket_name))
            cprint(cons.txt_gs_bucket_list_to_df_errors.format(bucket_name), 'yellow')
        else:
            cprint(cons.txt_gs_bucket_list_to_df_errors.format(bucket_name), 'yellow')


def storage_download_file(project_id, url=None,
                          logging=None, key_path=None):
    """
     :param project_id: String
     :param url: String
     :param logging: Object
     :param key_path: file.json
     :return: download_blobs
    """

    import os
    from urllib.parse import urlparse
    from color_tools import cprint
    from google.cloud import storage

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not url:
        raise Exception('require var url:{url} ')

    storage_client = storage.Client(project=project_id)
    if project_id and key_path:
        storage_client = storage.Client.from_service_account_json(key_path)

    def decode_gcs_url(url):
        p = urlparse(url)
        path = p.path[1:].split('/', 1)
        bucket, file_path = str(path[0]), path[1]
        return bucket, file_path

    bucket, file_path = decode_gcs_url(url)
    try:
        bucket = storage_client.bucket(bucket)
        blob = bucket.blob(file_path)
        blob.download_to_filename(os.path.basename(file_path))

        if logging:
            logging.info(cons.txt_bucket_download_file_success.format(os.path.basename(file_path)))
            cprint(cons.txt_bucket_download_file_success.format(os.path.basename(file_path)), 'yellow')
        else:
            cprint(cons.txt_bucket_download_file_success.format(os.path.basename(file_path)), 'yellow')

    except Exception as e:
        if logging:
            logging.info(cons.txt_bucket_download_file_errors.format(os.path.basename(file_path)))
            cprint(cons.txt_bucket_download_file_errors.format(os.path.basename(file_path)), 'yellow')
        else:
            cprint(cons.txt_bucket_download_file_errors.format(os.path.basename(file_path)), 'yellow')


def storage_upload_file(project_id, bucket_name=None,
                        destination_blob_name=None,
                        source_file_name=None,
                        is_public=False,
                        logging=None,
                        key_path=None):
    """
     :param project_id: String
     :param bucket_name: String
     :param destination_blob_name: String
     :param source_file_name: String
     :param logging: Object
     :param key_path: file.json
     :return: uploads_blobs
    """
    import os
    from color_tools import cprint
    from google.cloud import storage

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not bucket_name:
        raise Exception('require var bucket_name')
    if not destination_blob_name:
        raise Exception('require var destination_blob_name')
    if not source_file_name:
        raise Exception('require var source_file_name')

    storage_client = storage.Client(project=project_id)
    if project_id and key_path:
        storage_client = storage.Client.from_service_account_json(key_path)

    try:
        bucket = storage_client.bucket(bucket_name)
        blob = bucket.blob(destination_blob_name)
        blob.upload_from_filename(source_file_name)
        if is_public:
            blob.make_public()

        if logging:
            logging.info(cons.txt_bucket_upload_file_success.format(os.path.basename(source_file_name)))
            cprint(cons.txt_bucket_upload_file_success.format(os.path.basename(source_file_name)), 'yellow')
        else:
            cprint(cons.txt_bucket_upload_file_success.format(os.path.basename(source_file_name)), 'yellow')

    except Exception as e:
        if logging:
            logging.info(cons.txt_bucket_upload_file_errors.format(os.path.basename(source_file_name)))
            cprint(cons.txt_bucket_upload_file_errors.format(os.path.basename(source_file_name)), 'yellow')
        else:
            cprint(cons.txt_bucket_upload_file_errors.format(os.path.basename(source_file_name)), 'yellow')


def storage_delete_file(project_id,
                        bucket_name,
                        source_file_name=None,
                        logging=None,
                        key_path=None):
    """
     :param project_id: String
     :param bucket_name: String
     :param source_file_name: String
     :param logging: Object
     :param key_path: file.json
     :return: uploads_blobs
    """
    import os
    from color_tools import cprint
    from google.cloud import storage

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not bucket_name:
        raise Exception('require var bucket_name')
    if not source_file_name:
        raise Exception('require var source_file_name')

    storage_client = storage.Client(project=project_id)
    if project_id and key_path:
        storage_client = storage.Client.from_service_account_json(key_path)

    try:
        bucket = storage_client.bucket(bucket_name)
        blob = bucket.blob(source_file_name)
        blob.delete()

        if logging:
            logging.info(cons.txt_storage_delete_file_success.format(os.path.basename(source_file_name)))
            cprint(cons.txt_storage_delete_file_success.format(os.path.basename(source_file_name)), 'yellow')
        else:
            cprint(cons.txt_storage_delete_file_success.format(os.path.basename(source_file_name)), 'yellow')

    except Exception as e:
        if logging:
            logging.info(cons.txt_storage_delete_file_errors.format(os.path.basename(source_file_name)))
            cprint(cons.txt_storage_delete_file_errors.format(os.path.basename(source_file_name)), 'yellow')
        else:
            cprint(cons.txt_storage_delete_file_errors.format(os.path.basename(source_file_name)), 'yellow')


def storage_delete_folder(project_id=None,
                          bucket_name=None,
                          path_folder=None,
                          logging=None,
                          key_path=None):
    """
     :param bucket_name: The bucket name in which the file is to be placed
     :param path_folder: Folder name to be deleted
     :param logging: Object
     :param key_path: file.json
    :return: returns nothing
    """
    import os
    from color_tools import cprint
    from google.cloud import storage

    if not project_id:
        raise Exception('require var project_id:{project_id} ')
    if not bucket_name:
        raise Exception('require var bucket_name')
    if not path_folder:
        raise Exception('require var path_folder')

    storage_client = storage.Client(project=project_id)
    if project_id and key_path:
        storage_client = storage.Client.from_service_account_json(key_path)

    bucket = storage_client.bucket(bucket_name)
    try:
        bucket.delete_blobs(blobs=list(bucket.list_blobs(prefix=path_folder)))

        if logging:
            logging.info(cons.txt_storage_delete_folder_success.format(os.path.basename(path_folder)))
            cprint(cons.txt_storage_delete_folder_success.format(os.path.basename(path_folder)), 'yellow')
        else:
            cprint(cons.txt_storage_delete_folder_success.format(os.path.basename(path_folder)), 'yellow')

    except Exception as e:
        if logging:
            logging.info(cons.txt_storage_delete_folder_errors.format(os.path.basename(path_folder)))
            cprint(cons.txt_storage_delete_folder_errors.format(os.path.basename(path_folder)), 'yellow')
        else:
            cprint(cons.txt_storage_delete_folder_errors.format(os.path.basename(path_folder)), 'yellow')



def get_reduce_memory(dataset, verbose=True):
    import numpy as np
    import pandas as pd

    ds_tmp = dataset.copy()
    start_mem = ds_tmp.memory_usage().sum() / 1024 ** 2
    int_columns = ds_tmp.select_dtypes(include=[np.int8, np.int16, np.int32, np.int64]).columns.tolist()
    for col in int_columns:
        ds_tmp[col] = pd.to_numeric(arg=ds_tmp[col], downcast='integer')

    float_columns = ds_tmp.select_dtypes(include=[np.float32, np.float64]).columns.tolist()
    for col in float_columns:
        ds_tmp[col] = pd.to_numeric(arg=ds_tmp[col], downcast='float')

    end_mem = ds_tmp.memory_usage().sum() / 1024 ** 2
    ds_tmp = ds_tmp.replace([np.inf, -np.inf], np.nan)
    if verbose:
        print(f'Memory usage after optimization is: {end_mem:.2f} MB')
        print(f'Decreased by {100 * (start_mem - end_mem) / start_mem:.2f}%')
    del dataset
    return ds_tmp



import os
from distutils.core import setup
from setuptools import find_packages

setuppath = os.path.dirname(os.path.abspath(__file__))

setup(
    name='etl_bq_tools',
    packages=find_packages(),
    version='0.0.19',
    description='etl_bq_tools',
    long_description=open(os.path.join(setuppath, 'README.md')).read(),
    long_description_content_type="text/markdown",
    author='Jonathan Quiza',
    author_email='jony327@gmail.com',
    url='https://github.com/jonaqp/etl_bigquery_tools/',
    download_url='https://github.com/jonaqp/etl_bigquery_tools/archive/main.zip',
    keywords=['bigquery', 'etl', 'bq'],
    install_requires=open(os.path.join(setuppath, 'requirements.txt')).read().splitlines(),
    dependency_links=[],
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.7',
    ],
)



from etl_bq_tools.functions.bigquery import bq_create_dataset
from etl_bq_tools.functions.bigquery import bq_create_table
from etl_bq_tools.functions.bigquery import bq_delete_table
from etl_bq_tools.functions.bigquery import bq_delete_dataset
from etl_bq_tools.functions.bigquery import bq_list_dataset
from etl_bq_tools.functions.bigquery import bq_query_delete_partition
from etl_bq_tools.functions.bigquery import bq_query_to_df
from etl_bq_tools.functions.bigquery import bq_query_to_rows
from etl_bq_tools.functions.bigquery import bq_query_truncate
from etl_bq_tools.functions.bq import bq_execute_sql
from etl_bq_tools.functions.bq import bq_set_project
from etl_bq_tools.functions.cloud_storage import gs_bucket_list_patterns
from etl_bq_tools.functions.cloud_storage import gs_bucket_list_to_df
from etl_bq_tools.functions.cloud_storage import storage_download_file
from etl_bq_tools.functions.cloud_storage import storage_upload_file
from etl_bq_tools.functions.cloud_storage import storage_delete_file
from etl_bq_tools.functions.cloud_storage import storage_delete_folder

from etl_bq_tools.functions.dataframe import upload_df_to_bq
from etl_bq_tools.functions.dataset import show_describe_dataset
from etl_bq_tools.utils.logger import get_logger
from etl_bq_tools.utils.memory import get_reduce_memory
from etl_bq_tools.utils.prettytable import format_for_print
from etl_bq_tools.utils.time_execution import get_time_function_execution

bigquery_all = ["bq_create_dataset", "bq_create_table", "bq_delete_table",
                "bq_delete_dataset", "bq_list_dataset", "bq_query_to_df",
                "bq_query_to_rows", "bq_query_truncate", "bq_query_delete_partition"]

dataset_all = ["show_describe_dataset"]

bq_all = ["bq_execute_sql", "bq_set_project"]

cloud_storage_all = ["gs_bucket_list_patterns", "gs_bucket_list_to_df",
                     "storage_download_file", "storage_upload_file",
                     "storage_delete_file", "storage_delete_folder"]

dataframe_all = ["upload_df_to_bq"]

utils_all = ["get_logger", "get_reduce_memory", "get_time_function_execution",
             "format_for_print"]

__all__ = bigquery_all + dataset_all + bq_all + cloud_storage_all + dataframe_all + utils_all



import datetime
import time

from color_tools import cprint
from prettytable import PrettyTable


def get_stage(title, content):
    bcolors = dict(HEADER='\033[95m',
                   OKBLUE='\033[94m',
                   OKGREEN='\033[90m',
                   WARNING='\033[93m',
                   FAIL='\033[91m',
                   ENDC='\033[0m',
                   BOLD='\033[1m',
                   UNDERLINE='\033[4m')

    t = PrettyTable()
    t.field_names = [f'{bcolors.get("OKBLUE")}:::: {title} ::::{bcolors.get("ENDC")}']
    t.add_row([f'{bcolors.get("BOLD")} {content} {bcolors.get("ENDC")}'])
    print(t)


def get_time_function_execution(function_to_execute):
    def compute_execution_time(*args, **kwargs):
        start_time = time.time()
        result = function_to_execute(*args, **kwargs)
        if result is not None:
            end_time = time.time()
            tmin, tsec = divmod(datetime.timedelta(seconds=end_time - start_time).total_seconds(), 60)
            title = f"Time Execution: "
            content = f"{tmin:.0f} minutes and {round(tsec, 2)} seconds."
            cprint(f'{title} {content}', 'yellow')
        return result

    return compute_execution_time
